import path from 'path';
import fs from 'fs';
import { parse } from 'babylon';
import traverse from '@babel/traverse';
import generator from '@babel/generator';
import { StringLiteral } from '@babel/types';
import transform from 'cfx.babel';
import coffee from 'coffeescript';
import { createFilter } from 'rollup-pluginutils';

function formatError (source, filename, err) {
  var carets, col, first_column, first_line, l, last_column, last_line, lines;
  ({
    first_line,
    last_line,
    first_column,
    last_column
  } = err.location);

  if (last_line == null) {
    // last_line no longer exists randomly
    last_line = first_line;
  } // get lines from source


  lines = source.split('\n'); // insert start of highlighting

  l = lines[first_line];
  lines[first_line] = l.substring(0, first_column) + '\x1B[91m' + l.substring(first_column); // insert end

  l = lines[last_line];

  if (first_line === last_line) {
    // error only spans one line and overlaps with current cursor
    // skip a bit farther due to insertion of highlighting
    col = last_column + 6;
  } else {
    col = last_column;
  } // end highlighting


  lines[last_line] = l.substring(0, col) + '\x1B[39m' + l.substring(col); // get subset of lines with error to display

  lines = lines.slice(first_line, last_line + 1); // underline with carets

  carets = Array(first_column + 1).join(' ') + `\x1B[91m${Array(last_column + 2 - first_column).join('^')}\x1B[39m`;
  lines.push(carets);
  return `${filename}:${first_line + 1}:${first_column + 1}: ${err.name}: ${err.message}\n${lines.join('\n')}\n`;
}

// import dd from 'ddeyes'
var ASTToCode, getAST, rCFPath, ricffAST;

rCFPath = filePath => {
  var fileObj, getFileObj;

  getFileObj = file => {
    return {
      exists: fs.existsSync(file),
      ...path.parse(file)
    };
  };

  fileObj = getFileObj(filePath);

  if (!fileObj.exists) {
    // file isnt exists
    fileObj = {
      coffee: (() => {
        var _path;

        _path = path.format({
          root: path.join(fileObj.root, fileObj.dir + '/'),
          name: fileObj.base,
          ext: '.coffee'
        });
        return {
          path: _path,
          obj: getFileObj(_path)
        };
      })(),
      js: (() => {
        var _path;

        _path = path.format({
          root: path.join(fileObj.root, fileObj.dir + '/'),
          name: fileObj.base,
          ext: '.js'
        });
        return {
          path: _path,
          obj: getFileObj(_path)
        };
      })()
    };

    if (fileObj.coffee.obj.exists) {
      return rCFPath(fileObj.coffee.path);
    }

    if (fileObj.js.obj.exists) {
      return rCFPath(fileObj.js.path);
    }

    return null; // file exists
  } else {
    if (fileObj.ext !== '') {
      // isnt dir
      return filePath;
    } else {
      return rCFPath(path.join(filePath, 'index.coffee'));
    }
  }
}; // replace import coffee file from AST


ricffAST = (ast, dirname = __dirname) => {
  traverse(ast, {
    ImportDeclaration: _ast => {
      var currentNode, fileParseObj, filePath;
      currentNode = _ast.node;
      fileParseObj = path.parse(currentNode.source.value);

      if (fileParseObj.root === '' && fileParseObj.dir === '') {
        return;
      }

      filePath = rCFPath(path.join(dirname, currentNode.source.value));

      if (filePath != null) {
        _ast.node.source = StringLiteral(filePath);
      }
    }
  });
  return ast;
};

getAST = source => {
  return parse(source, {
    sourceType: 'module',
    plugins: ['objectRestSpread', 'asyncGenerators']
  });
};

ASTToCode = (ast, es = true, opts) => {
  if (es === false) {
    return transform(ast, opts);
  } else {
    return generator(ast).code;
  }
};

// import dd from 'ddeyes'
var sourceMap;

sourceMap = out => {
  if (out.v3SourceMap) {
    return JSON.parse(out.v3SourceMap);
  } else {
    return null;
  }
};

var coffee2 = ((opts = {}) => {
  var filter;

  if (opts.bare == null) {
    opts.bare = true;
  }

  if (opts.sourceMap == null) {
    opts.sourceMap = true;
  }

  filter = createFilter(opts.include, opts.exclude);
  return {
    name: 'coffee2',
    transform: (code, id) => {
      var _code, err, out;

      if (!filter(id)) {
        return {
          code
        };
      }

      if (path.extname(id) !== '.coffee') {
        return {
          code
        };
      }

      try {
        out = coffee.compile(code, {
          filename: id,
          bare: opts.bare,
          sourceMap: opts.sourceMap
        });
      } catch (error) {
        err = error;

        if (err.location != null) {
          err.formattedMessage = formatError(code, id, err);
        }

        throw err;
      }

      _code = ASTToCode(ricffAST(getAST(out.js), path.dirname(id)));
      return {
        code: _code,
        map: sourceMap(out)
      };
    }
  };
});

// import dd from 'ddeyes'
var babel = ((opts = {}) => {
  var filter;
  filter = createFilter(opts.include, opts.exclude);
  return {
    name: 'babel',
    transform: (code, id) => {
      var _code;

      if (!filter(id)) {
        return {
          code
        };
      }

      if (path.extname(id) !== '.js') {
        return {
          code
        };
      }

      _code = ASTToCode(getAST(code), false, opts.opts);
      return _code;
    }
  };
});

export { coffee2, babel };
